from django import forms
from django.contrib import admin
from django.utils.translation import gettext_lazy as _

from django_admin_keycloak.models import KeycloakProvider

admin.site.login_template = 'django_admin_keycloak/login.html'


class KeycloakProviderModelForm(forms.ModelForm):
    class Meta:
        model = KeycloakProvider
        fields = '__all__'
        widgets = {
            'client_secret': forms.PasswordInput()
        }


@admin.register(KeycloakProvider)
class KeycloakProviderAdmin(admin.ModelAdmin):
    form = KeycloakProviderModelForm
    list_display = ('name', 'slug', 'active',)
    list_display_links = ('name',)

    def get_fieldsets(self, request, obj=None):
        server_fields = [
            'server_url',
            'realm_name',
            'client_id',
            'client_secret',
            'scope',
        ]
        if obj and getattr(obj, 'is_static', False):
            server_fields.remove('client_secret')
        return (
            (None, {'fields': (
                'active',
                'slug',
                'name',
            )}),
            (_('Server'), {'fields': server_fields}),
            (_('Other Info'), {'fields': (
                'redirect_uri',
                'role_super_user',
                'role_staff_user',
                'options',
            )})
        )

    def get_exclude(self, request, obj=None):
        excluded = list(super().get_exclude(request, obj) or [])
        if obj and getattr(obj, 'is_static', False):
            return excluded.append('client_secret')
        return excluded

    def has_change_permission(self, request, obj=None):
        if obj and getattr(obj, 'is_static', False):
            return False
        return super().has_change_permission(request, obj)

    def has_delete_permission(self, request, obj=None):
        if obj and getattr(obj, 'is_static', False):
            return False
        return super().has_delete_permission(request, obj)

    search_fields = ('name', 'slug')



from django import template
from django_admin_keycloak.models import KeycloakProvider
from django_admin_keycloak.oidc import get_auth_url

register = template.Library()


@register.simple_tag(takes_context=True)
def keycloak_authorization_url(context):
    request = context['request']
    return get_auth_url(request)


def _get_providers(request):
    for provider in KeycloakProvider.objects.filter(active=True).iterator():
        yield {
            'name': provider.name,
            'url': get_auth_url(provider, request)
        }


@register.inclusion_tag('django_admin_keycloak/auth_link.html', takes_context=True)
def keycloak_authorization_links(context):
    request = context['request']
    count = KeycloakProvider.objects.filter(active=True).count()
    return {
        'providers': _get_providers(request),
        'has_one': count == 1,
        'has_many': count > 1
    }


@register.simple_tag(name='sso_account', takes_context=True)
def get_account_link(context):
    request = context['request']
    try:
        provider = KeycloakProvider.objects.get(pk=request.session['keycloak']['pk'])
    except (KeycloakProvider.DoesNotExist, KeyError):
        return

    return provider.get_account_link()



from django.apps import AppConfig
from django.core.checks import register


class DjangoAdminKeycloakConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "django_admin_keycloak"
    verbose_name = "Keycloak"

    def ready(self):
        from django_admin_keycloak.static_settings import check_static_config
        register(check_static_config, 'keycloak', name="keycloak_static_config", )



from django.urls import path

from .views import auth_callback, auth_logout

urlpatterns = [
    path('callback/<slug:keycloak_slug>', auth_callback, name='oidc-login'),
    path('sso-logout', auth_logout, name='oidc-logout'),
    path('sso-error', auth_logout, name='oidc-login-error'),
]



import base64
import json
import logging

from django.db.transaction import atomic
from django.http import Http404, JsonResponse
from django.http import HttpResponseRedirect
from django.shortcuts import redirect, render
from django.views.decorators.csrf import csrf_exempt
from keycloak import KeycloakPostError

from .models import KeycloakProvider, KeycloakSession
from .oidc import auth_oidc

logger = logging.getLogger('django_admin_keycloak')


def auth_callback(request, keycloak_slug: str):
    try:
        provider = KeycloakProvider.objects.get(slug=keycloak_slug)
    except KeycloakProvider.DoesNotExist:
        raise Http404("Not Found")
    try:
        redirect_uri = auth_oidc(
            provider=provider,
            request=request
        )
    except KeycloakPostError as e:
        logger.error(e)
        return redirect('django_admin_keycloak:oidc-login-error')

    return HttpResponseRedirect(redirect_uri)


def auth_error(request):
    return render(request, 'django_admin_keycloak/login-error.html')


@csrf_exempt
def auth_logout(request):
    if not hasattr(request.session, 'sid'):
        return JsonResponse({'success': True})
    token = request.body.decode('utf-8').split('=', maxsplit=1)[1]
    base64_string = token.split('.')[1]
    data = json.loads(base64.b64decode(f'{base64_string}====').decode('utf-8'))
    sid = data.get('sid')
    with atomic():
        KeycloakSession.objects.filter(sid=sid).delete()

    return JsonResponse({'success': True})



# Generated by Django 4.2.8 on 2023-12-27 12:31

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("django_admin_keycloak", "0001_initial"),
    ]

    operations = [
        migrations.RenameField(
            model_name="keycloakprovider",
            old_name="client_secret_key",
            new_name="client_secret",
        ),
        migrations.AddField(
            model_name="keycloakprovider",
            name="is_static",
            field=models.BooleanField(
                default=False, editable=False, verbose_name="is static"
            ),
        ),
        migrations.AddField(
            model_name="keycloakprovider",
            name="scope",
            field=models.TextField(default="openid profile", verbose_name="scope"),
        ),
    ]



from django.conf import settings
from django.core.checks import Error

PROVIDERS = getattr(settings, 'KEYCLOAK_PROVIDERS', None)

REQUIRED_PROPERTIES = (
    'slug',
    'name',
    'server_url',
    'realm_name',
    'client_id',
    'client_secret',
    'redirect_uri',
    'role_super_user',
    'role_staff_user',
)


def check_static_config(app_configs, **kwargs):
    errors = []
    if PROVIDERS is None:
        return errors

    for provider in PROVIDERS:
        for rq in REQUIRED_PROPERTIES:
            if (rq in provider and provider[rq] is None) or rq not in provider:
                errors.append(Error(f'Missing required property {rq}'))
    return errors



from django.contrib.sessions.backends.db import SessionStore as DBStore

from django_admin_keycloak.models import KeycloakSession


class SessionStore(DBStore):

    def __init__(self, session_key=None):
        super().__init__(session_key)
        self._sid = None

    @classmethod
    def get_model_class(cls):
        return KeycloakSession

    def create_model_instance(self, data):
        obj = super().create_model_instance(data)
        obj.sid = self._sid
        return obj

    @property
    def sid(self):
        return self._sid

    @sid.setter
    def sid(self, value):
        self._sid = value





# Generated by Django 5.0 on 2023-12-18 20:45

import django_admin_keycloak.models
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="KeycloakProvider",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("active", models.BooleanField(default=True, verbose_name="active")),
                ("slug", models.SlugField(unique=True, verbose_name="slug")),
                ("name", models.CharField(max_length=128, verbose_name="name")),
                ("server_url", models.URLField(verbose_name="server url")),
                (
                    "realm_name",
                    models.CharField(max_length=128, verbose_name="realm name"),
                ),
                (
                    "client_id",
                    models.CharField(max_length=128, verbose_name="client id"),
                ),
                (
                    "client_secret_key",
                    models.CharField(
                        blank=True,
                        max_length=128,
                        null=True,
                        verbose_name="client secret key",
                    ),
                ),
                (
                    "redirect_uri",
                    models.CharField(
                        default="/",
                        help_text="The redirect URI for user is not staff",
                        max_length=256,
                        verbose_name="redirect_uri",
                    ),
                ),
                (
                    "role_super_user",
                    models.CharField(
                        default="superuser",
                        max_length=128,
                        verbose_name="superuser role",
                    ),
                ),
                (
                    "role_staff_user",
                    models.CharField(
                        default="staff", max_length=128, verbose_name="staff user role"
                    ),
                ),
                (
                    "options",
                    models.JSONField(
                        default=django_admin_keycloak.models._default_options,
                        help_text="Available options list see https://python-keycloak.readthedocs.io/",
                        verbose_name="options",
                    ),
                ),
            ],
            options={
                "verbose_name": "Keycloak Provider",
                "verbose_name_plural": "Keycloak Provider",
                "db_table": "keycloak_provider",
            },
        ),
    ]



from datetime import timedelta

from django.contrib.auth import get_user_model, login
from django.contrib.auth.models import Group, User
from django.http import HttpRequest
from django.urls import reverse
from keycloak import KeycloakOpenID

from django_admin_keycloak.models import KeycloakProvider

UserModel = get_user_model()


def get_oidc_client_by_pk(pk: int) -> KeycloakOpenID | None:
    try:
        provider = KeycloakProvider.objects.get(pk=pk)
    except KeycloakProvider.DoesNotExist:
        return
    return get_oidc_client(provider)


def get_oidc_client_by_slug(slug: str) -> KeycloakOpenID | None:
    try:
        provider = KeycloakProvider.objects.get(slug=slug)
    except KeycloakProvider.DoesNotExist:
        return
    return get_oidc_client(provider)


def get_oidc_client(provider: KeycloakProvider) -> KeycloakOpenID:
    return KeycloakOpenID(
        server_url=provider.server_url,
        realm_name=provider.realm_name,
        client_id=provider.client_id,
        client_secret_key=provider.client_secret,
        **provider.options
    )


def get_auth_url(provider: KeycloakProvider, request: HttpRequest, state: str | None = None) -> str:
    oidc_client = get_oidc_client(provider)
    return oidc_client.auth_url(
        request.build_absolute_uri(reverse('oidc-login', kwargs={'keycloak_slug': provider.slug})),
        scope=provider.scope,
        state=state or ''
    )


def get_next_url(request: HttpRequest, default: str) -> str:
    try:
        next_url = request.GET['next']
        if isinstance(next_url, (list, tuple)):
            next_url = next_url[0]
        return next_url
    except KeyError:
        return default


def auth_oidc(provider: KeycloakProvider, request: HttpRequest):
    oidc_client = get_oidc_client(provider)
    code = request.GET.get('code')
    sid = request.GET.get('session_state')
    access_token = oidc_client.token(
        grant_type='authorization_code',
        code=code,
        redirect_uri=request.build_absolute_uri(reverse('oidc-login', kwargs={'keycloak_slug': provider.slug}))
    )

    request.session['keycloak'] = {
        'sid': sid,
        'token': access_token,
        'pk': provider.pk
    }
    request.session.sid = sid

    userinfo = oidc_client.userinfo(token=access_token['access_token'])

    user = create_user(userinfo)
    check_permissions(provider, user, userinfo)
    check_member_of(provider, user, userinfo)
    login(request, user)
    request.session.set_expiry(timedelta(seconds=access_token['refresh_expires_in']))

    return get_next_url(request, provider.redirect_uri)


def create_user(userinfo: dict) -> UserModel:
    username = userinfo['preferred_username']
    try:
        user = UserModel._default_manager.get_by_natural_key(username)
    except UserModel.DoesNotExist:
        user = UserModel.objects.create_user(
            username=username,
            email=userinfo['email'],
            first_name=userinfo['given_name'],
            last_name=userinfo['family_name']
        )
    return user


def check_permissions(provider: KeycloakProvider, user: UserModel, userinfo: dict):
    groups = userinfo.get('groups')
    if not groups:
        return

    is_superuser = provider.role_super_user in groups
    is_staff = provider.role_staff_user in groups
    changed = False
    if user.is_superuser != is_superuser:
        user.is_superuser = is_superuser
        changed = True
    if user.is_staff != is_staff:
        user.is_staff = is_staff
        changed = True
    if changed:
        user.save(update_fields=['is_superuser', 'is_staff'])


def check_member_of(provider: KeycloakProvider, user: UserModel, userinfo: dict):
    groups = userinfo.get('groups')
    if not groups:
        return
    user.groups.clear()
    user.groups.add(*Group.objects.filter(name__in=groups))



from django.contrib.auth import get_user_model
from django.contrib.sessions.base_session import AbstractBaseSession
from django.contrib.sessions.models import SessionManager
from django.db import models
from django.utils.translation import gettext_lazy as _

UserModel = get_user_model()


def _default_options() -> dict:
    return {
        'verify': True,
        'custom_headers': None,
        'proxies': None,
        'timeout': 60
    }


class KeycloakProvider(models.Model):
    active = models.BooleanField(_('active'), default=True)
    slug = models.SlugField(_('slug'), unique=True)
    name = models.CharField(_('name'), max_length=128)

    server_url = models.URLField(_('server url'))
    realm_name = models.CharField(_('realm name'), max_length=128)
    client_id = models.CharField(_('client id'), max_length=128)
    client_secret = models.CharField(_('client secret key'), max_length=128, blank=True, null=True)
    scope = models.TextField(_('scope'), default='openid profile')

    redirect_uri = models.CharField(
        _('redirect_uri'),
        max_length=256,
        default='/',
        help_text=_('The redirect URI for user is not staff')
    )
    role_super_user = models.CharField(_('superuser role'), max_length=128, default='superuser')
    role_staff_user = models.CharField(_('staff user role'), max_length=128, default='staff')

    options = models.JSONField(
        _('options'),
        default=_default_options,
        help_text=_('Available options list see https://python-keycloak.readthedocs.io/')
    )
    is_static = models.BooleanField(_('is static'), default=False, editable=False)

    def __str__(self):
        return self.name or self.realm_name

    def get_account_link(self) -> str:
        return f'{self.server_url}/realms/{self.realm_name}/account'

    class Meta:
        db_table = 'keycloak_provider'
        verbose_name = _('Keycloak Provider')
        verbose_name_plural = _('Keycloak Provider')


class KeycloakSession(AbstractBaseSession):
    objects = SessionManager()

    sid = models.UUIDField(null=True)

    @classmethod
    def get_session_store_class(cls):
        from django.contrib.sessions.backends.db import SessionStore

        return SessionStore

    class Meta(AbstractBaseSession.Meta):
        db_table = "keycloak_session"



from django.core.management import BaseCommand

from django_admin_keycloak.models import KeycloakProvider
from django_admin_keycloak.static_settings import PROVIDERS


class Command(BaseCommand):
    help = 'Updates static Keycloak providers'

    @staticmethod
    def remove_static_providers(*slugs):
        qs = KeycloakProvider.objects.exclude(slug__in=slugs).filter(is_static=True)
        names = list(qs.values_list('name', flat=True))

        qs.delete()
        return names

    @staticmethod
    def create_or_update_provider(provider: dict) -> tuple['KeycloakProvider', bool]:
        provider.setdefault('options', {
            'verify': True,
            'custom_headers': None,
            'proxies': None,
            'timeout': 60
        })
        provider.setdefault('redirect_uri', '/')
        provider.setdefault('active', True)
        provider['is_static'] = True
        allowed_field = [f.name for f in KeycloakProvider._meta.get_fields()]
        defaults = {}
        for k, v in provider.items():
            if k in allowed_field:
                defaults[k] = v

        return KeycloakProvider.objects.update_or_create(defaults=defaults, slug=defaults['slug'])

    def handle(self, *args, **options):
        providers = PROVIDERS or []

        for provider in providers:
            obj, created = self.create_or_update_provider(provider)
            self.stdout.write(f'- {"Updating" if not created else "Created"} provider {obj}\n')

        for name in self.remove_static_providers(*[provider['slug'] for provider in providers]):
            self.stdout.write(f'- Remove provider {name}\n')



# Generated by Django 4.2.8 on 2024-01-04 09:15

import django.contrib.sessions.models
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        (
            "django_admin_keycloak",
            "0002_rename_client_secret_key_keycloakprovider_client_secret_and_more",
        ),
    ]

    operations = [
        migrations.CreateModel(
            name="KeycloakSession",
            fields=[
                (
                    "session_key",
                    models.CharField(
                        max_length=40,
                        primary_key=True,
                        serialize=False,
                        verbose_name="session key",
                    ),
                ),
                ("session_data", models.TextField(verbose_name="session data")),
                (
                    "expire_date",
                    models.DateTimeField(db_index=True, verbose_name="expire date"),
                ),
                ("sid", models.UUIDField(null=True)),
            ],
            options={
                "verbose_name": "session",
                "verbose_name_plural": "sessions",
                "db_table": "keycloak_session",
                "abstract": False,
            },
            managers=[
                ("objects", django.contrib.sessions.models.SessionManager()),
            ],
        ),
    ]
